/*
 * Package : Coap
 * Author : S. Hamblett <steve.hamblett@linux.com>
 * Date   : 12/04/2017
 * Copyright :  S.Hamblett
 */

import 'dart:math';

import 'package:typed_data/typed_data.dart';

import 'coap_option.dart';

/// This class describes the block options of the CoAP messages
class CoapBlockOption extends CoapOption {
  /// Base construction
  CoapBlockOption(super.type) {
    intValue = 0;
  }

  /// num - Block number
  /// szx - Block size
  /// m - More flag
  CoapBlockOption.fromParts(
    super.type,
    final int num,
    final int szx, {
    final bool m = false,
  }) {
    intValue = _encode(num, szx, m);
  }

  /// Sets block params.
  /// num - Block number
  /// szx - Block size
  /// m - More flag
  void setValue(final int num, final int szx, {required final bool m}) {
    intValue = _encode(num, szx, m);
  }

  /// The raw value
  set rawValue(final int num) => intValue = num;

  int get rawValue => num;

  /// Block number.
  int get num => intValue >> 4;

  set num(final int num) => setValue(num, szx, m: m);

  /// Block size.
  int get szx => intValue & 0x7;

  set szx(final int szx) => setValue(num, szx, m: m);

  /// More flag.
  bool get m => (intValue >> 3 & 0x1) != 0;

  set m(final bool m) => setValue(num, szx, m: m);

  /// Block bytes
  Uint8Buffer get blockValueBytes => _compressValueBytes();

  /// Gets the real block size which is 2 ^ (SZX + 4).
  static int decodeSZX(final int szx) => 1 << (szx + 4);

  /// Gets the decoded block size in bytes (B).
  int size() => decodeSZX(szx);

  /// Converts a block size into the corresponding SZX.
  static int encodeSZX(final int blockSize) {
    if (blockSize < 16) {
      return 0;
    }
    if (blockSize > 1024) {
      return 6;
    }
    return ((log(blockSize) / log(2)) - 4).toInt();
  }

  /// Checks whether the given SZX is valid or not.
  static bool validSZX(final int szx) => szx >= 0 && szx <= 6;

  @override
  String toString() => 'Raw value: $intValue, num: $num, szx: $szx, more: $m';

  static int _encode(final int num, final int szx, final bool m) {
    var value = 0;
    value |= szx & 0x7;
    value |= (m ? 1 : 0) << 3;
    value |= num << 4;
    return value;
  }

  /// Strips leading zeros for 32 bit integers
  Uint8Buffer _compressValueBytes() {
    if (byteValue.length == 4) {
      if (byteValue[3] == 0) {
        return Uint8Buffer()..addAll(byteValue.take(3).toList());
      }
    }
    return byteValue;
  }
}
